/*-----------------------------------------------------------------*/
/*! 
  \file calceph.c 
  \brief high-level API 
         access and interpolate INPOP and JPL Ephemeris data.
         compute position and velocity from binary ephemeris file.

  \author  M. Gastineau 
           Astronomie et Systemes Dynamiques, IMCCE, CNRS, Observatoire de Paris. 

   Copyright, 2008, 2009, M. Gastineau, IMCCE-CNRS
   email of the author : gastineau@imcce.fr

  History:                                                                
    \bug M. Gastineau 22/07/08 : support of TT-TDB for INPOP
    \note M. Gastineau 30/01/09 : add functions calceph_(s)getconstantcount 
          and calceph_(s)getconstantindex
*/
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
/* License  of this file :
 This file is "dual-licensed", you have to choose one  of the two licenses 
 below to apply on this file.
 
    CeCILL-C
    	The CeCILL-C license is close to the GNU LGPL.
    	( http://www.cecill.info/licences/Licence_CeCILL-C_V1-en.html )
 
 or CeCILL v2.0
      The CeCILL license is compatible with the GNU GPL.
      ( http://www.cecill.info/licences/Licence_CeCILL_V2-en.html )
 

This library is governed by the CeCILL-C or the CeCILL license under 
French law and abiding by the rules of distribution of free software.  
You can  use, modify and/ or redistribute the software under the terms 
of the CeCILL-C or CeCILL license as circulated by CEA, CNRS and INRIA  
at the following URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C or CeCILL license and that you accept its terms.
*/
/*-----------------------------------------------------------------*/

#include "calcephconfig.h"
#if HAVE_STDIO_H
#include <stdio.h>
#endif
#if HAVE_LIMITS_H
#include <limits.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_PTHREAD
#include <pthread.h>
#endif /*HAVE_PTHREAD*/
#define __CALCEPH_WITHIN_CALCEPH 1
#include "calceph.h"
#include "real.h"
#include "calcephinternal.h"
#include "util.h"

/********************************************************/
/* local variables */
/********************************************************/
#ifdef HAVE_PTHREAD
/* version thread-safe */

/*! key to access pephbin structure for thread-safe */
static pthread_key_t calceph_key;
/*! key to run only one time calcephinit_once (thread-safe) */
static pthread_once_t calceph_key_once = PTHREAD_ONCE_INIT;

static void calceph_key_init(void);

#else /*HAVE_PTHREAD*/

/* version not thread-safe */
/*! description of the ephemeris file  */
static t_calcephbin *pephbin = NULL;
#endif /*HAVE_PTHREAD*/

/********************************************************/
/*! Open the specified ephemeris file 
  return 0 on error.
  return 1 on success.

 @param filename (in) nom du fichier 
*/
/********************************************************/
int calceph_sopen(const char *filename)
{
#ifdef HAVE_PTHREAD
 /*thread-safe */
 t_calcephbin *pephbin = NULL;
 pthread_once(&calceph_key_once, calceph_key_init);
 
 pephbin = calceph_open(filename);
 
 if (pthread_setspecific(calceph_key, pephbin)!=0)
     fatalerror("Can't specify thread-data for calceph_sopen.\n");  
#else /*HAVE_PTHREAD*/
  pephbin = calceph_open(filename);
#endif /*HAVE_PTHREAD*/
  return (pephbin!=NULL);
}

/********************************************************/
/*! Close the  ephemeris file 
*/
/********************************************************/
void calceph_sclose()
{
#ifdef HAVE_PTHREAD
 /*thread-safe */
 t_calcephbin *pephbin;
 pephbin = (t_calcephbin *)pthread_getspecific(calceph_key);
#endif /*HAVE_PTHREAD*/
 if (pephbin==NULL)
 { /*first call => initialize for this thread */
  fatalerror("calceph_sopen must be called before calceph_sclose (for each thread).\n");
 }
 else
 {
  calceph_close(pephbin);
 }
}

/********************************************************/
/*! Return the position and the velocity of the object 
  "target" in the vector PV at the specified time 
  (time ellapsed from JD0).
  The center is specified in center.
  
  return 0 on error.
  return 1 on success.

   @param JD0 (in) reference time (could be 0)
   @param time (in) time ellapsed from JD0
   @param target (in) "target" object 
   @param center (in) "center" object 
   @param PV (out) position and the velocity of the "target" object
*/ 
/********************************************************/
int calceph_scompute(treal JD0, treal time, int target, int center, treal PV[6])
{
 int res = 0;
#ifdef HAVE_PTHREAD
 /*thread-safe */
 t_calcephbin *pephbin;
 pephbin = (t_calcephbin *)pthread_getspecific(calceph_key);
#endif /*HAVE_PTHREAD*/
 if (pephbin==NULL)
 { /*first call => initialize for this thread */
  fatalerror("calceph_sopen must be called before calceph_scompute (for each thread).\n");
 }
 else
 {
  res = calceph_compute(pephbin, JD0, time, target, center, PV);
 }
 return res;
}
 

/********************************************************/
/*! return the value of the specified constant from the ephemeris file 
  return 0 if constant is not found
  return 1 if constant is found and set value.
  
  @param name (in) constant name
  @param value (out) value of the constant
*/
/********************************************************/
int calceph_sgetconstant(const char* name, double *value)
{
 int res = 0;
 
#ifdef HAVE_PTHREAD
 /*thread-safe */
 t_calcephbin *pephbin;
 pephbin = (t_calcephbin *)pthread_getspecific(calceph_key);
#endif /*HAVE_PTHREAD*/
 if (pephbin==NULL)
 { /*first call => initialize for this thread */
  fatalerror("calceph_sopen must be called before calceph_sgetconstant (for each thread).\n");
 }
 else
 {
  res = calceph_getconstant(pephbin, name, value);
 }
 return res;
}

/********************************************************/
/*! return the number of constants available in the ephemeris file 
  
*/
/********************************************************/
int calceph_sgetconstantcount(void)
{
 int res = 0;
 
#ifdef HAVE_PTHREAD
 /*thread-safe */
 t_calcephbin *pephbin;
 pephbin = (t_calcephbin *)pthread_getspecific(calceph_key);
#endif /*HAVE_PTHREAD*/
 if (pephbin==NULL)
 { /*first call => initialize for this thread */
  fatalerror("calceph_sopen must be called before calceph_sgetconstantcount (for each thread).\n");
 }
 else
 {
  res = calceph_getconstantcount(pephbin);
 }
 return res;
}

/********************************************************/
/*! get the name and its value of the constant available 
    at some index from the ephemeris file 
   return 1 if index is valid
   return 0 if index isn't valid
  
  @param index (in) index of the constant (must be between 1 and calceph_getconstantcount() )
  @param name (out) name of the constant 
  @param value (out) value of the constant 
*/
/********************************************************/
int calceph_sgetconstantindex(int index, char name[CALCEPH_MAX_CONSTANTNAME],
                             double *value)
{
 int res = 0;
 
#ifdef HAVE_PTHREAD
 /*thread-safe */
 t_calcephbin *pephbin;
 pephbin = (t_calcephbin *)pthread_getspecific(calceph_key);
#endif /*HAVE_PTHREAD*/
 if (pephbin==NULL)
 { /*first call => initialize for this thread */
  fatalerror("calceph_sopen must be called before calceph_sgetconstantindex (for each thread).\n");
 }
 else
 {
  res = calceph_getconstantindex(pephbin,index,name, value);
 }
 return res;
}


/********************************************************/
/*! Return the position and the velocity of the object 
  "target" in the vector PV at the specified time 
  (time ellapsed from JD0).
  The center is specified in center.
  
  return 0 on error.
  return 1 on success.

   @param pephbin (inout) ephemeris file descriptor
   @param JD0 (in) reference time (could be 0)
   @param time (in) time ellapsed from JD0
   @param target (in) "target" object 
   @param center (in) "center" object 
   @param PV (out) position and the velocity of the "target" object
*/ 
/********************************************************/
int calceph_compute(t_calcephbin* pephbin, treal JD0, treal time, int target, int center, treal PV[6])
{
 stateType  statetarget,statecenter, statelune;
 int detarget, decenter, delune;
 double EMRAT;
 int res = 0;
 int j;
 
 EMRAT = Ephemeris_getEMRAT(pephbin);
 target--;
 center--;
 
 delune   = MOON;
 /* cas lune ou terre : toujours recuperer la lune */
 if (target==MOON || center==MOON || center==EARTH || target==EARTH)
 {
  res = calceph_interpol_PV(pephbin, JD0, time,  delune,  &statelune);
 }
 
 /* cas vecteur Terre-Lune */
 if (center==EARTH  && target==MOON)
 {
  for (j=0;j<3; j++) 
  {
   PV[j]=statelune.Position[j];
   PV[j+3]=statelune.Velocity[j];
  }
 }
 /* cas vecteur Lune-Terre */
 else if (target==EARTH && center==MOON)
 {
  for (j=0;j<3; j++) 
  {
   PV[j]=-statelune.Position[j];
   PV[j+3]=-statelune.Velocity[j];
  }
 }
 /*cas libration  ou TT-TDB */
 else if ((target==LIBRATIONS || target==TTMTDB) && center==-1)
 {
  res = calceph_interpol_PV(pephbin, JD0, time, target , &statetarget);
  for (j=0;j<3; j++) 
  {
   PV[j]=statetarget.Position[j];
   PV[j+3]=statetarget.Velocity[j];
  }
 }
 /* autre cas */
 else
 {
  detarget = target;
  decenter = center;
  if (target==EM_BARY || target==MOON) detarget=EARTH;
  if (decenter==EM_BARY || center==MOON) decenter=EARTH;
  for (j=0;j<3; j++) 
  {
   PV[j]=0.E0;
   PV[j+3]=0.E0;
  }
  
  if (detarget!=SS_BARY && detarget!=-1)
  {
   res = calceph_interpol_PV(pephbin, JD0, time, detarget , &statetarget);
   for (j=0;j<3; j++) 
   {
    PV[j] = statetarget.Position[j];
    PV[j+3] = statetarget.Velocity[j];
   }
   /* cas target = MOON */
   if (target==MOON)
   {
    for (j=0;j<3; j++) 
    {
     PV[j] += statelune.Position[j]*EMRAT/(REALC(1.0E0)+EMRAT); 
     PV[j+3] += statelune.Velocity[j]*EMRAT/(REALC(1.0E0)+EMRAT); 
    }
   }
   /* cas target == EARTH */
   if (target==EARTH)
   {
    for (j=0;j<3; j++) 
    {
     PV[j] -= statelune.Position[j]/(REALC(1.0E0)+EMRAT); 
     PV[j+3] -= statelune.Velocity[j]/(REALC(1.0E0)+EMRAT); 
    }
   }
  }
  
  if (decenter!=SS_BARY && decenter!=-1)
  {
   res = calceph_interpol_PV(pephbin, JD0, time, decenter , &statecenter);
   for (j=0;j<3; j++) 
   {
    PV[j] -= statecenter.Position[j];
    PV[j+3] -= statecenter.Velocity[j];
   }
   /* cas center = MOON */
   if (center==MOON)
   {
    for (j=0;j<3; j++) 
    {
     PV[j] -= statelune.Position[j]*EMRAT/(REALC(1.0E0)+EMRAT); 
     PV[j+3] -= statelune.Velocity[j]*EMRAT/(REALC(1.0E0)+EMRAT); 
    }
   }
   /* cas center == EARTH */
   if (center==EARTH)
   {
    for (j=0;j<3; j++) 
    {
     PV[j] += statelune.Position[j]/(REALC(1.0E0)+EMRAT); 
     PV[j+3] += statelune.Velocity[j]/(REALC(1.0E0)+EMRAT); 
    }
   }
  }
 }
 return res;
}

/********************************************************/
/*! return the value of the specified constant from the ephemeris file 
  return 0 if constant is not found
  return 1 if constant is found and set value.
  
  @param eph (inout) ephemeris descriptor
  @param name (in) constant name
  @param value (out) value of the constant
*/
/********************************************************/
int calceph_getconstant(t_calcephbin* eph, const char* name, double *value)
{
 int res = 0;
 if (strcmp(name,"AU")==0)
 {
    *value = Ephemeris_getAU(eph);
    res = 1;
 }
 else if (strcmp(name,"EMRAT")==0)
 {
    *value = Ephemeris_getEMRAT(eph);
    res = 1;
 }
 else
 {
    res = Ephemeris_getConstant(eph, name, value);
 }
 return res;
}



#ifdef HAVE_PTHREAD
/********************************************************/
/*! 
    Initialize calceph_key for thread-safe usage\n
    Must be called only by fcntabtcheb (once time )
*/
/********************************************************/
static void calceph_key_init(void)
{
 if (pthread_key_create(&calceph_key,NULL)!=0)
  fatalerror("Can't initialize thread-key for calceph_init.\n");
}
#endif /*HAVE_PTHREAD*/
