/*-----------------------------------------------------------------*/
/*! 
  \file calcephio.c 
  \brief perform I/O on ephemeris data file 
         interpolate INPOP and JPL Ephemeris data.
         compute position and velocity from binary ephemeris file.

  \author  M. Gastineau 
           Astronomie et Systemes Dynamiques, IMCCE, CNRS, Observatoire de Paris. 

   Copyright, 2008, 2009,  M. Gastineau, IMCCE-CNRS
   email of the author : gastineau@imcce.fr

  History:                                                                
 \bug M. Gastineau 01/03/05 : thread-safe version
 \bug M. Gastineau 09/09/05 : support of the libration
 \bug M. Gastineau 18/11/05 : incrase the number of coefficients
 \bug M. Gastineau 22/03/06 : support of non-native "endian" files
 \bug M. Gastineau 11/05/06 : support of DE411 and DE414
 \bug M. Gastineau 13/11/06 : support of km or UA units for  INPOP
 \bug M. Gastineau 22/07/08 : support of TT-TDB for INPOP
 \bug M. Gastineau 17/10/08 : support of DE421
*/
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
/* License  of this file :
 This file is "dual-licensed", you have to choose one  of the two licenses 
 below to apply on this file.
 
    CeCILL-C
    	The CeCILL-C license is close to the GNU LGPL.
    	( http://www.cecill.info/licences/Licence_CeCILL-C_V1-en.html )
 
 or CeCILL v2.0
      The CeCILL license is compatible with the GNU GPL.
      ( http://www.cecill.info/licences/Licence_CeCILL_V2-en.html )
 

This library is governed by the CeCILL-C or the CeCILL license under 
French law and abiding by the rules of distribution of free software.  
You can  use, modify and/ or redistribute the software under the terms 
of the CeCILL-C or CeCILL license as circulated by CEA, CNRS and INRIA  
at the following URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C or CeCILL license and that you accept its terms.
*/
/*-----------------------------------------------------------------*/

#include "calcephconfig.h"
#if HAVE_STDIO_H
#include <stdio.h>
#endif
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#if HAVE_ERRNO_H
#include <errno.h>
#endif
#if HAVE_MATH_H
#include <math.h>
#endif
#ifndef __cplusplus
#if HAVE_STDBOOL_H
#include <stdbool.h>
#endif
#endif /*__cplusplus */

#include "real.h"
#define __CALCEPH_WITHIN_CALCEPH 1
#include "calceph.h"
#include "calcephinternal.h"
#include "util.h"

/**==========================================================================**/
/**  Structure globale                                                       **/
/**==========================================================================**/
/*! descripteur de l'ephemeride */
struct calcephbin {
   t_HeaderBlock   H1;          /*!< contenu du 1er record du fichier */
   FILE        *Ephemeris_File; /*!< fichier contenant l'ephemeride */
   double       *Coeff_Array ;  /*!< tableau des coefficients */
   double       T_beg;          /*!< temps du debut du record contenu dans Coeff_Array */
   double       T_end;          /*!< temps de la fin du record contenu dans Coeff_Array */
   double       T_span;         /*!< temps d'un granule dans Coeff_Array */
   int          ARRAY_SIZE;     /*!< nombre de coefficients contenus dans un record */ 
   int          ncomp;          /*!< nombre de composante danes le fichier */
   double       constVal[400];  /*!< valeur des constantes */
   bool         swapbyteorder;  /*!< = true : swap the bytes after each read operation */
   bool         isinAU;         /*!< = true : coefficients are in UA (else in km) */
   bool         haveTTmTDB;     /*!< = ture : TT-TDB est dans le fichier */
};
  
#define DEBUG 0             /*!< =0 => no debug output  */

/*==========================================================================*/
/* Fonctions locales                                                        */
/*==========================================================================*/

static bool calceph_ReadCoeff(t_calcephbin *pephbin,  double Time );
static void calceph_interpol_PV_an(t_calcephbin *p_pbinfile,  treal TimeJD0 , 
                                 treal Timediff, stateType *Planet, long C, long G, long N);

/*--------------------------------------------------------------------------*/
/*!                                                                       
     This function is used by the functions below to read an array of     
     Tchebeychev coefficients from a binary ephemeris data file.          
  @return return true if time is available 
  
  @param pephbin(inout) descriptor of the ephemerids
  @param Time (in) Desired record time.   
  
*/
/*--------------------------------------------------------------------------*/
static bool calceph_ReadCoeff(t_calcephbin *pephbin,  double Time )
{
  double  T_delta = 0.0;
  int     Offset  =  0 ;

  /*--------------------------------------------------------------------------*/
  /*  Find ephemeris data that record contains input time. Note that one, and */
  /*  only one, of the following conditional statements will be true (if both */
  /*  were false, this function would not have been called).                  */
  /*--------------------------------------------------------------------------*/

  if ( Time < pephbin->T_beg )        /* Compute backwards location offset */
     {
       T_delta = pephbin->T_beg - Time;
       Offset  = (int) -ceil(T_delta/pephbin->T_span); 
     }

  if ( Time > pephbin->T_end )        /* Compute forwards location offset */
     {
       T_delta = Time - pephbin->T_end;
       Offset  = (int) ceil(T_delta/pephbin->T_span);
     }

  /*--------------------------------------------------------------------------*/
  /*  Retrieve ephemeris data from new record.                                */
  /*--------------------------------------------------------------------------*/

  fseek(pephbin->Ephemeris_File,(Offset-1)*pephbin->ARRAY_SIZE*sizeof(double),
        SEEK_CUR);
  fread(pephbin->Coeff_Array,sizeof(double),pephbin->ARRAY_SIZE,
        pephbin->Ephemeris_File);
  /* swap the data if necessary */
  if (pephbin->swapbyteorder) swapdblarray(pephbin->Coeff_Array, pephbin->ARRAY_SIZE);
  
  pephbin->T_beg  = pephbin->Coeff_Array[0];
  pephbin->T_end  = pephbin->Coeff_Array[1];
  pephbin->T_span = pephbin->T_end - pephbin->T_beg;


  /* verifie s'il y a eu une erreur de fread ou de fseek */
  if ( Time < pephbin->T_beg || Time>pephbin->T_end) 
     {
       fatalerror("time %g isn't in the interval [%.16g , %.16g] \n",
                  Time, pephbin->H1.timeData[0], pephbin->H1.timeData[1]);
       return false;
     }
  return true;
}

/*--------------------------------------------------------------------------*/
/*!  
     Open the "filename" ephemeris file .
     Read the header data block
     Read the constant data block
     Read the first coefficient data bock.
     
  @return descriptor of the ephemeris and the first array of coefficients   
  
  @param fileName (in) A character string giving the name of an ephemeris data file.   
*/
/*--------------------------------------------------------------------------*/
t_calcephbin *calceph_open(const char *fileName )
{
  t_calcephbin *res;
  
  /* allocate memory */
  res = (t_calcephbin *) malloc(sizeof(t_calcephbin));
  if (res==NULL)
  {
   fatalerror("Can't allocate memory for t_calcephbin\nSystem error : '%s'\n",
              strerror(errno));

   return NULL;
  }
  
  /*--------------------------------------------------------------------------*/
  /*  Open ephemeris file.                                                    */
  /*--------------------------------------------------------------------------*/
   
  res->Ephemeris_File = fopen(fileName,"r");

  /*--------------------------------------------------------------------------*/
  /*  Read header & first coefficient array, then return status code.         */
  /*--------------------------------------------------------------------------*/

  if (res->Ephemeris_File == NULL ) 
     {
      fatalerror("Unable to open ephemeris file: '%s'\nSystem error : '%s'\n",
                 fileName, strerror(errno));
      free(res);
      res = NULL;
     }
  else 
     { /*Read first three header records from ephemeris file */
         
       fread(&res->H1,sizeof(t_HeaderBlock),1,res->Ephemeris_File);
       res->ncomp = 3;
       res->swapbyteorder = false;
       res->isinAU = false;
       res->haveTTmTDB = false;
       switch (res->H1.DENUM)
       {
        case EPHEMERIS_DE200: res->ARRAY_SIZE = ARRAY_SIZE_DE200; break;
        case EPHEMERIS_DE403: res->ARRAY_SIZE = ARRAY_SIZE_DE403; break;
        case EPHEMERIS_DE405: res->ARRAY_SIZE = ARRAY_SIZE_DE405; break;
        case EPHEMERIS_DE406: res->ARRAY_SIZE = ARRAY_SIZE_DE406; break;
        case EPHEMERIS_DE411: res->ARRAY_SIZE = ARRAY_SIZE_DE411; break;
        case EPHEMERIS_DE414: res->ARRAY_SIZE = ARRAY_SIZE_DE414; break;
        case EPHEMERIS_DE418: res->ARRAY_SIZE = ARRAY_SIZE_DE418; break;
        case EPHEMERIS_DE421: res->ARRAY_SIZE = ARRAY_SIZE_DE421; break;
 
        case EPHEMERIS_INPOP: 
                      res->ARRAY_SIZE = res->H1.recordsize; 
                      res->ncomp = 6;
                      break;
        default:
          switch (swapint(res->H1.DENUM))
          {
          case EPHEMERIS_DE200: res->ARRAY_SIZE = ARRAY_SIZE_DE200; res->swapbyteorder = true; break;
          case EPHEMERIS_DE403: res->ARRAY_SIZE = ARRAY_SIZE_DE403; res->swapbyteorder = true; break;
          case EPHEMERIS_DE405: res->ARRAY_SIZE = ARRAY_SIZE_DE405; res->swapbyteorder = true; break;
          case EPHEMERIS_DE406: res->ARRAY_SIZE = ARRAY_SIZE_DE406; res->swapbyteorder = true; break;
          case EPHEMERIS_DE411: res->ARRAY_SIZE = ARRAY_SIZE_DE411; res->swapbyteorder = true; break;
          case EPHEMERIS_DE414: res->ARRAY_SIZE = ARRAY_SIZE_DE414; res->swapbyteorder = true; break;
          case EPHEMERIS_DE418: res->ARRAY_SIZE = ARRAY_SIZE_DE418; res->swapbyteorder = true; break;
          case EPHEMERIS_DE421: res->ARRAY_SIZE = ARRAY_SIZE_DE421; res->swapbyteorder = true; break;
          case EPHEMERIS_INPOP: 
                      res->swapbyteorder = true; 
                      res->ARRAY_SIZE = res->H1.recordsize; 
                      if (res->swapbyteorder) res->ARRAY_SIZE = swapint(res->ARRAY_SIZE);
                      res->ncomp = 6;
                      break;
          default:
            fatalerror("Opened wrong file: '%s' for ephemeris: %d.\n",
                     fileName, res->H1.DENUM);
            return NULL;
            break;
          }
          break;
       }
       
      /*--------------------------------------------------------------------------*/
      /* swap the header if necessary */
      /*--------------------------------------------------------------------------*/
      if (res->swapbyteorder)
      {
        swapdblarray(res->H1.timeData,3);
        res->H1.numConst = swapint(res->H1.numConst);
        res->H1.AU = swapdbl(res->H1.AU);
        res->H1.EMRAT = swapdbl(res->H1.EMRAT);
        swapintarray((int*)res->H1.coeffPtr,36);
        res->H1.DENUM = swapint(res->H1.DENUM);
        swapintarray(res->H1.libratPtr,3);
        swapintarray(res->H1.TTmTDBPtr,3);
        res->H1.recordsize = swapint(res->H1.recordsize);     
      }
       
      /*--------------------------------------------------------------------------*/
      /* read constants */
      /*--------------------------------------------------------------------------*/
       res->Coeff_Array = (double *) malloc(sizeof(double)*res->ARRAY_SIZE);
       if (res->Coeff_Array==NULL)
       {
        fatalerror("Can't allocate memory for %d real\nSystem error : '%s'\n",
                   res->ARRAY_SIZE, strerror(errno));
        return NULL;

       }
       /*lecture des constantes */      
       /*se positionne au debut du deuxieme record */
       fseek(res->Ephemeris_File,sizeof(double)*res->ARRAY_SIZE,SEEK_SET);
       fread(res->constVal,sizeof(double),400, res->Ephemeris_File);
       /* swap the data if necessary */
       if (res->swapbyteorder) swapdblarray(res->constVal, 400);
             
       /* ajuste le nombre de composantes pour inpop */
       if (res->H1.DENUM == EPHEMERIS_INPOP)
       {
        double dformat;
        double dunite;
        if (Ephemeris_getConstant(res,"FORMAT",&dformat))
        {
         if (((int)(fmod(dformat,10)))==1) { res->ncomp = 3;  };
         if (((int)(fmod(rint(dformat/10),10)))==1){ res->haveTTmTDB = true;  };
        }
        if (Ephemeris_getConstant(res,"UNITE",&dunite))
        {
         res->isinAU = (((int)dunite)==0);
        }
        else
        {
         fatalerror("Can't find constant 'UNITE' in ephemeris file. Wrong version of ephemeris file\n");
         return NULL;
        }
       }
       /*se positionne au debut du troisieme record */
       fseek(res->Ephemeris_File,2*sizeof(double)*res->ARRAY_SIZE,SEEK_SET);
       fread(res->Coeff_Array,sizeof(double),res->ARRAY_SIZE,
             res->Ephemeris_File);
       /* swap the data if necessary */
       if (res->swapbyteorder) swapdblarray(res->Coeff_Array, res->ARRAY_SIZE);
       
       /* current interval */

       res->T_beg  = res->Coeff_Array[0];
       res->T_end  = res->Coeff_Array[1];
       res->T_span = res->T_end - res->T_beg;

 
#if DEBUG
          {
            printf("\n  calceph_open : \n");
            printf("\n      ARRAY_SIZE = %4d",res->ARRAY_SIZE);
            printf("\n      T_Beg      = %7.3f",res->T_beg);
            printf("\n      T_End      = %7.3f",res->T_end);
            printf("\n      T_Span     = %7.3f",res->T_span);
            printf("\n      Number of component     = %d",res->ncomp);
            printf("\n      TT-TDB     = %d",res->haveTTmTDB);
            printf("\n      record size     = %d\n\n",res->H1.recordsize);
          }
#endif /*DEBUG*/

     }
 return res;
}

/*--------------------------------------------------------------------------*/
/*! Close the  ephemeris file and destroy the ephemeris descriptor
 @param eph (inout) ephemeris descriptor to be destroyed
*/
/*--------------------------------------------------------------------------*/
void calceph_close(t_calcephbin* eph)
{
 fclose(eph->Ephemeris_File);
 free(eph->Coeff_Array);
 free(eph);
}

/*--------------------------------------------------------------------------*/
/*!                                                      
     This function read the data block corresponding to the specified Time
     if this block isn't already in memory.
     
     It computes position and velocity vectors for a selected  
     planetary body from Chebyshev coefficients.          

                                                                          
  @param p_pbinfile (inout) descripteur de l'ephemeride                            
  @param TimeJD0+timediff  (in) Time for which position is desired (Julian Date)
  @param Target   (in) Solar system body for which position is desired.         
  @param Position (in) Pointer to external array to receive the position.       
  @param Planet (out) position and veloticity                                                                        
         Planet has position in ua                                                
         Planet has velocity in ua/day
*/
/*--------------------------------------------------------------------------*/

int calceph_interpol_PV(t_calcephbin *p_pbinfile,  treal TimeJD0 , treal Timediff,
                        int Target, stateType *Planet)
{
  treal    Time = TimeJD0+Timediff;
  int       i ;
  long int  C , G , N ;


  if ( Target==NUTATIONS)
     {
       fatalerror("\n This function does not compute nutations.\n");
       return 0;
     }

  /*--------------------------------------------------------------------------*/
  /* Determine if a new record needs to be input.                             */
  /*--------------------------------------------------------------------------*/
  
  if (Time < p_pbinfile->T_beg || Time > p_pbinfile->T_end)  
  {
    if (!calceph_ReadCoeff(p_pbinfile,Time)) return 0;
  }
  
  /*--------------------------------------------------------------------------*/
  /* Read the coefficients from the binary record.                            */
  /*--------------------------------------------------------------------------*/
  
  if ( Target == LIBRATIONS )   
  { /*Librations */
   C = p_pbinfile->H1.libratPtr[0] - 1;  
   N = p_pbinfile->H1.libratPtr[1];      
   G = p_pbinfile->H1.libratPtr[2];       
  }
  else if ( Target == TTMTDB )   
  { /*TT-TDB */
   if (!p_pbinfile->haveTTmTDB) 
   {
    fatalerror(" The file doesn't have TT-TDB data.\n"); 
    return 0;
   }
   C = p_pbinfile->H1.TTmTDBPtr[0] - 1;  
   N = p_pbinfile->H1.TTmTDBPtr[1];      
   G = p_pbinfile->H1.TTmTDBPtr[2];      
  }
  else
  {/* Planets */
   C = p_pbinfile->H1.coeffPtr[Target][0] - 1;  
   N = p_pbinfile->H1.coeffPtr[Target][1];      
   G = p_pbinfile->H1.coeffPtr[Target][2];       
  }
  
  calceph_interpol_PV_an(p_pbinfile,  TimeJD0 ,  Timediff, Planet,C, G, N);

 
  /* only for planets */
  if (Target<11 && !p_pbinfile->isinAU)
   {
     treal rau = p_pbinfile->H1.AU;
     /* pour DExxx, conversion de km en UA */
     for ( i=0 ; i<3 ; i++ )
     {
       Planet->Position[i] /= rau;
       Planet->Velocity[i] /= rau;
     }
   }
  
  return 1;
}

/*--------------------------------------------------------------------------*/
/*!  This function computes position and velocity vectors                 
     from Chebyshev coefficients These coefficients are stored at offset  
     C, G in  the object p_pbinfile.                                      
                                          
                                                                          
 @param p_pbinfile (inout) ephemeris descriptor                            
 @param TimeJD0+timediff  (in) Time for which position is desired (Julian Date)
 @param Planet (out) Pointer to external array to receive the position/velocity.
   Planet has position in ephemerides units                                 
   Planet has velocity in ephemerides units                                 
 @param C (inout)   Coeff array entry point                                       
 @param N(in)    Number of coeff's                                           
 @param G (in)   Granules in current record                                    
*/
/*--------------------------------------------------------------------------*/
static void calceph_interpol_PV_an(t_calcephbin *p_pbinfile,  treal TimeJD0 , 
                        treal Timediff, stateType *Planet, 
                        long C, long G, long N)
{
  treal    A[15000]   ,  Cp[15000] , P_Sum[3] , V_Sum[3] , Up[15000] ,
            T_break , T_sub  , Tc;
  treal    Time = TimeJD0+Timediff;
  long long int TimeJD0pent = longr(TimeJD0);
  treal TimeJD0frac = TimeJD0-realr(TimeJD0pent);
  long long int Timediffpent = longr(Timediff);
  treal Timedifffrac = Timediff-realr(Timediffpent);
  long long int T_begpent;
  treal T_begfrac;
  treal deltapent , deltafrac;
  int       i , j;
  long int  offset = 0;
  stateType X;

 
  /*--------------------------------------------------------------------------*/
  /* Initialize local coefficient array.                                      */
  /*--------------------------------------------------------------------------*/

  for ( i=0 ; i<15000 ; i++ )
      {
        A[i] = REALC(0.0E0);
      }


  T_begpent = longr(p_pbinfile->T_beg);
  T_begfrac = p_pbinfile->T_beg-T_begpent;
  if ( G == 1 )
     {
      /* code equivalent a :
       Tc = 2.0E0*(Time - T_beg) / T_span - 1.0E0; */
      deltapent = (TimeJD0pent + Timediffpent - T_begpent) / p_pbinfile->T_span;
      deltafrac = (TimeJD0frac + Timedifffrac - T_begfrac) / p_pbinfile->T_span;
     }
     else
     {
       T_sub = p_pbinfile->T_span / realr(G);  /* Compute subgranule interval */
       
       for ( j=G ; j>0 ; j-- ) 
           {
             T_break = p_pbinfile->T_beg + realr(j-1) * T_sub;
             if ( Time > T_break ) 
                {
                 /* T_seg  = T_break;*/
                  offset = j-1;
                  break;
                }
            }
      /* code equivalent a :
         Tc = 2.0E0*(Time - T_seg) / T_sub - 1.0E0; */
       deltapent = (TimeJD0pent + Timediffpent - T_begpent) / T_sub;
       deltafrac = (TimeJD0frac + Timedifffrac - T_begfrac) / T_sub
                   - realr(offset);    
       
       C  = C + p_pbinfile->ncomp * offset * N;       
     }
     Tc = REALC(2.0E0)*(deltapent+deltafrac) - REALC(1.0E0);
     for (i=C ; i<(C+p_pbinfile->ncomp*N) ; i++) 
         A[i-C] = p_pbinfile->Coeff_Array[i];


  /*--------------------------------------------------------------------------*/
  /* Compute the interpolated position and velocity                           */
  /*--------------------------------------------------------------------------*/
  
  if (p_pbinfile->ncomp==3)
  {
   for ( i=0 ; i<3 ; i++ )                /* Compute interpolating polynomials */
      {
        Cp[0] = REALC(1.0E0);           
        Cp[1] = Tc;
        Cp[2] = REALC(2.0E0) * Tc*Tc - REALC(1.E0);
        
        Up[0] = REALC(0.0E0);
        Up[1] = REALC(1.0E0);
        Up[2] = REALC(4.0E0) * Tc;

        for ( j=3 ; j<N ; j++ )
            {
              Cp[j] = REALC(2.0E0) * Tc * Cp[j-1] - Cp[j-2];
              Up[j] = REALC(2.0E0) * Tc * Up[j-1] + REALC(2.0E0) * Cp[j-1] - Up[j-2];
            }

        P_Sum[i] = REALC(0.0E0);           /* Compute interpolated position & velocity */
        V_Sum[i] = REALC(0.0E0);

        for ( j=N-1 ; j>-1 ; j-- )  P_Sum[i] = P_Sum[i] + A[j+i*N] * Cp[j];
        for ( j=N-1 ; j>0  ; j-- )  V_Sum[i] = V_Sum[i] + A[j+i*N] * Up[j];

        X.Position[i] = P_Sum[i];
        X.Velocity[i] = V_Sum[i] * REALC(2.0E0) * realr(G) / (p_pbinfile->T_span);
      }
   }
   else
   {
     for ( i=0 ; i<3 ; i++ )  /* Compute interpolating polynomials */
      {
        Cp[0] = REALC(1.0E0);           
        Cp[1] = Tc;
        Cp[2] = REALC(2.0E0) * Tc*Tc - REALC(1.E0);
        
        for ( j=3 ; j<N ; j++ )
            {
              Cp[j] = REALC(2.0E0) * Tc * Cp[j-1] - Cp[j-2];
            }

        P_Sum[i] = REALC(0.0E0);           /* Compute interpolated position & velocity */
        V_Sum[i] = REALC(0.0E0);

        for ( j=N-1 ; j>-1 ; j-- )  
        {
         P_Sum[i] = P_Sum[i] + A[j+i*N] * Cp[j];
         V_Sum[i] = V_Sum[i] + A[j+(3+i)*N] * Cp[j];
        }

        X.Position[i] = P_Sum[i];
        X.Velocity[i] = V_Sum[i];
      }
   }
   
  /*--------------------------------------------------------------------------*/
  /*  Return computed values.                                                 */
  /*--------------------------------------------------------------------------*/

  *Planet = X;

}

/*--------------------------------------------------------------------------*/
/*! Open the specified ephemeris file 
 @param filename (in) nom du fichier 
*/
/*--------------------------------------------------------------------------*/
void calcephinit(const char *filename)
{
 calceph_sopen(filename);
}

/*--------------------------------------------------------------------------*/
/*! Return the position and the velocity of the object 
  "target" in the vector PV at the specified time 
  (time ellapsed from JD0).
  The center is specified in center.
   @param JD0 (in) reference time (could be 0)
   @param time (in) time ellapsed from JD0
   @param target (in) "target" object 
   @param center (in) "center" object 
   @param PV (out) position and the velocity of the "target" object
*/ 
/*--------------------------------------------------------------------------*/
void calceph(treal JD0, treal time, int target, int center, treal PV[6])
{
 calceph_scompute(JD0, time, target, center, PV);
}


/*--------------------------------------------------------------------------*/
/*! return EMRAT from file 
  @param p_pbinfile (in) ephemeris file
*/
/*--------------------------------------------------------------------------*/
double Ephemeris_getEMRAT(t_calcephbin *p_pbinfile)
{
 return p_pbinfile->H1.EMRAT;
}

/*--------------------------------------------------------------------------*/
/*! return AU from file 
  @param p_pbinfile (in) ephemeris file
*/
/*--------------------------------------------------------------------------*/
double Ephemeris_getAU(t_calcephbin *p_pbinfile)
{
 return p_pbinfile->H1.AU;
}


/*--------------------------------------------------------------------------*/
/*! return the value of the specified constant from file 
  return 0 if constant is not found
  return 1 if constant is found and set p_pdval.
  @param p_pbinfile (in) ephemeris file
  @param name (in) constant name
  @param p_pdval (out) value of the constant
*/
/*--------------------------------------------------------------------------*/
int Ephemeris_getConstant(t_calcephbin *p_pbinfile, 
                           const char *name, double *p_pdval)
{
 int found = 0;
 int cpt=0;
 size_t len = strlen(name);
 while (cpt<400 && found==0)
 {
  if (strncmp(name, (const char*)(p_pbinfile->H1.constName[cpt]),len)==0)
  {
   found = 1;
   *p_pdval = p_pbinfile->constVal[cpt];
  }
  cpt++;
 }
 return found;
}

/*--------------------------------------------------------------------------*/
/*! return the number of constants available in the ephemeris file 
  
  @param eph (inout) ephemeris descriptor
*/
/*--------------------------------------------------------------------------*/
int calceph_getconstantcount(t_calcephbin* eph)
{
 return eph->H1.numConst+2; /*+2 <=> AU + EMRAT */ 
}

/*--------------------------------------------------------------------------*/
/*! get the name and its value of the constant available 
    at some index from the ephemeris file 
   return 1 if index is valid
   return 0 if index isn't valid
  
  @param eph (inout) ephemeris descriptor
  @param index (in) index of the constant (must be between 1 and calceph_getconstantcount() )
  @param name (out) name of the constant 
  @param value (out) value of the constant 
*/
/*--------------------------------------------------------------------------*/
int calceph_getconstantindex(t_calcephbin* eph, int index, char name[CALCEPH_MAX_CONSTANTNAME],
                             double *value)
{
 int res;
 int j;
 int nconst = calceph_getconstantcount(eph);
 
 if (index>=1 && index<=nconst)
 {
  res = 1;
  for (j=0; j<CALCEPH_MAX_CONSTANTNAME; j++) name[j]=' '; /* initialization for fortran */
  if (index==nconst)
  {
   *value = Ephemeris_getAU(eph);
   strcpy(name,"AU");
  }
  else if (index==nconst-1)
  {
   *value = Ephemeris_getEMRAT(eph);
   strcpy(name,"EMRAT");
  }
  else
  {
    strncpy(name, eph->H1.constName[index-1], CALCEPH_MAX_CONSTANTNAME-1);
    *value = eph->constVal[index-1];
  }
  name[CALCEPH_MAX_CONSTANTNAME-1] ='\0';
 }
 else
 {
  res = 0;
 }
 return res;
}

